//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright Contributors to the OpenEXR Project.
//

// clang-format off

#ifndef _PyImathFun_h_
#define _PyImathFun_h_

#include "PyImathExport.h"

namespace PyImath {

PYIMATH_EXPORT void register_functions();

}

#endif
